!|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

      module ccsm3_grid

!***********************************************************************
!
!     Based upon pop grid.F  CVS:$Name: ccsm3_0_rel04 $
!
!-----------------------------------------------------------------------

      use ccsm3                   ! ccsm3 ocean
      use ccsm3_constants         ! ccsm3 ocean constants
      use ccsm3_io                ! ccsm3 ocean io
      use ccsm3_exit              ! ccsm3 ocean exit 
      use ccsm3_global_reductions ! ccsm3 mpi communications 
      use mod_dimensions          ! hycom module
      use mod_xc                  ! hycom module
      

      implicit none
      save


!-----------------------------------------------------------------------
!
!     geometric scalars
!
!-----------------------------------------------------------------------
      real*4, dimension(itdm,jtdm) :: real24
      real*4, dimension(4,itdm,jtdm) :: real44

      integer, dimension(itdm,jtdm) ::
     &   IMSK_G                         ! global ocean mask

      real, public, dimension(imt1:imt2,jmt1:jmt2) ::
     &  TLONG                   !
     &, TLAT
     &, TAREA                   !T-cell area 
     &, ANGLET                   
     &, RCALCT                  !mask (1. or 0.)



!***********************************************************************

      contains

!***********************************************************************

      subroutine ccsm3_grid_init (gridpath)

!-----------------------------------------------------------------------
!
!     intializes all grid quantities
!
!-----------------------------------------------------------------------
      character (*) :: gridpath


!-----------------------------------------------------------------------
!
!     local variables
!
!-----------------------------------------------------------------------

      character*80  :: gridname, gridtype

      namelist /coupled_grid_nml/gridname, gridtype

      integer ::
     &  nml_error           ! namelist i/o error flag
     &, i,j

 
!-----------------------------------------------------------------------
!
!     read input namelist for grid setup options
!
!-----------------------------------------------------------------------

      gridname      = 'unknown_horiz_gridname'
      gridtype      = 'unknown_horiz_gridtype'

      if (my_task == master_task) then
        nml_error = -1
        open (nml_in, file=nml_filename, status='old')
   10   continue  !*** keep reading until find right namelist
        read(nml_in, nml=coupled_grid_nml,err=10,end=20)
        close(nml_in)
        nml_error = 0
   20   continue
      endif

      call broadcast_scalar(nml_error, master_task)
      if (nml_error /= 0) then
        call ccsm3_exit_HYCOM ('ERROR reading grid_nml')
      endif

      if (my_task == master_task) then
        write(stdout,1000)
        write(stdout,1001)
        write(stdout,1002)
        write(stdout,1001)
        write(stdout,1000)
        call flush(stdout)
 1000   format(72('-'))
 1001   format(' ')
 1002   format(' Grid options')

        write(stdout,*) ' '
        write(stdout,*) ' Document Namelist Parameters:'
        write(stdout,*) ' ============================ '
        write(stdout,*) ' '
        write(stdout, coupled_grid_nml)
        write(stdout,*) ' '
        call flush(stdout)
      endif

      call broadcast_scalar(gridname, master_task)
      call broadcast_scalar(gridtype, master_task)

!-----------------------------------------------------------------------
!
!     read horizontal grid
!
!-----------------------------------------------------------------------

#if defined(gx3v5) || defined(gx1v3)
      ! with gx initialization, use the grid defined in geopar 
      call ccsm3_grid_set_horiz
#else
      ! with GLB2x initialization, grid is defined separately
      call ccsm3_grid_read_horiz (gridpath, gridname, gridtype)
#endif

      if (my_task == master_task) then
        write(stdout,*) ' '
        write(stdout,*) ' (ccsm3_grid_init) completed '
        call flush(stdout)
      endif


      end subroutine ccsm3_grid_init


!***********************************************************************

      subroutine ccsm3_grid_read_horiz (gridpath, gridname, gridtype)

!-----------------------------------------------------------------------
!
!     read horizontal grid information from input grid file
!
!     Comments from David Bi:
!     coupler needs following grid information (in 'normal' x-y directions):
!     TLON, TLAT, ANGLET: T cell lon, lat, and displaced angle in radians;
!     TAREA:              T cell area in square radians;
!     IMASK_G:            T point land mask (1-ocean, 0-land);
!
!     They could be obtained in geopar but currently geopar only read 
!     plon and plat (ie, tlon and tlat). Here we read all these grid
!     point data in init_coupled. 
!     (NOTE: we must pre-prepare the gridinfo datafile gridinfo_PanAm21.bin
!     somewhere else in advance!) 
!..........................................................................

!
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
!
!     input variables
!
!-----------------------------------------------------------------------

      character (*), intent(in) ::
     &  gridpath            ! file path name for horiz_grid_file
     &, gridname            ! filename of file containing grid data
     &, gridtype            ! grid type, eg, 'composite', or 'displaced_pole'
!-----------------------------------------------------------------------
!
!     local variables
!
!-----------------------------------------------------------------------

      real, dimension(itdm,jtdm) ::
     &   RMSK_G,                        ! global ocean mask (real)
     &   TLAT_G,                        ! latitude  of cell center
     &   TLON_G,                        ! longitude of cell center
     &   AREA_G                         ! T cell area in radian^2
     &  ,bchk

      real*4, dimension  (itdm,jtdm) :: real24
      real*4, dimension(4,itdm,jtdm) :: real44

      integer ::
     &  nu                    ! i/o unit number

!-----------------------------------------------------------------------

      if (my_task == master_task) then
        write(stdout,*) ' '
        write(stdout,*) ' file read: ', gridname
        call flush(stdout)

        open(101, file=trim(gridpath)//trim(gridname),
     &            form='unformatted',status='old')

        read(101) real24
        TLAT_G = real24
        read(101) real24
        TLON_G = real24
        read(101) real24
       !TANG_G = real24           !do not assign; not required by cpl6
        read(101) real44
       !TLAT_V = real44           !do not assign; not required by cpl6
        read(101) real44          
       !TLON_V = real44           !do not assign; not required by cpl6
        read(101) real24
        AREA_G = real24
        read(101) real24
        IMSK_G = int(real24)
        close(101)
      endif

      !DBI: assigning RCALCT with 1. or 0.
      where (ip /= 0)
        RCALCT = c1p
      elsewhere
        RCALCT = c0p
      endwhere
 
!-----------------------------------------------------------------------
!    Comments by David Bi:
!    NOTE: any global variables defined only on an individual processor 
!    (ie the master_task) must be boradcasted onto other processors 
!    before collective operaions are condtcued such as "call xcaput". 
!    Otherwise the result would be wrong 'cos the processor DON'T know 
!    the global variable and give a default value as, say, 0.
!    (So in recv_from_cpl, all arrays from cpl are initially received
!    and stored on the master_task, and thus MUST BE broadcasted onto
!    all processors....   (a silly ignorance had caused me big trouble!)
!-----------------------------------------------------------------------

 
!-----------------------------------------------------------------------
!
!     scatter global arrays prior to communicating with cpl6
!
!-----------------------------------------------------------------------
      call broadcast_vector_real(TLON_G, master_task, itdm*jtdm)
      call xcaput(TLON_G, TLONG, 0)
      call xctilr(TLONG,1,1,nbdy,nbdy,halo_ps)
 
      call broadcast_vector_real(TLAT_G, master_task, itdm*jtdm)
      call xcaput(TLAT_G, TLAT, 0)
      call xctilr(TLAT,1,1,nbdy,nbdy,halo_ps)

      call broadcast_vector_real(AREA_G, master_task, itdm*jtdm)
      call xcaput(AREA_G, TAREA, 0)
      call xctilr(TAREA,1,1,nbdy,nbdy,halo_ps)

      call broadcast_vector_int (IMSK_G, master_task, itdm*jtdm)
      call xciaput(IMSK_G,IMSK  , 0)
!     call xctilr(IMSK,1,1,nbdy,nbdy,halo_ps)  ! no halo update for integer arrays


      if (my_task == master_task) then
        write(stdout,*) ' '
        write(stdout,*) ' (ccsm3_grid_read_horiz) completed '
        call flush(stdout)
      endif

!-----------------------------------------------------------------------

      end subroutine ccsm3_grid_read_horiz


      subroutine ccsm3_grid_set_horiz

      integer :: i,j,n

!-----------------------------------------------------------------------
!
!     set horizontal grid information for cpl6, based upon plat, plon
!
!     cpl6 needs the following grid information (in 'normal' x-y directions):
!     TLON, TLAT: T cell lon, lat
!     TAREA:      T cell area in square radians;
!     IMASK_G:    T point land mask (1-ocean, 0-land);
!
!..........................................................................

      
      do j=1,jj
      do i=1,ii
         TLONG(i,j) = plon(i,j) 
         TLAT (i,j) = plat(i,j)
         TAREA(i,j) = scp2(i,j)*1.0e4  ! convert from m^2  to cm^2
         IMSK (i,j) = ip  (i,j)   
      enddo
      enddo


      !DBI: assigning RCALCT with 1. or 0.
      where (ip /= 0)
        RCALCT = c1p
      elsewhere
        RCALCT = c0p
      endwhere

      end subroutine ccsm3_grid_set_horiz



      end module ccsm3_grid

!|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
