c --- -----------------------------------------
c --- MICOM-based HYbrid Coordinate Ocean Model
c ---               H Y C O M
c ---           v e r s i o n  2.1
c --- -----------------------------------------

      program hycom

      use mod_xc      ! HYCOM communication interface
                      ! includes common_blocks.h via ccsm3 usage
      use mod_za      ! HYCOM I/O interface
      use mod_pipe    ! HYCOM debugging interface
      use mod_geopar  ! module containing subroutine geopar

      use ccsm3, only : ccsm3_setup_coupling_env ! ccsm3 ocean
      use ccsm3_exit                             ! ccsm3 ocean exit
      use ccsm3_io                               ! ccsm3 ocean io
      use ccsm3_forcing                          ! ccsm3 ocean/cpl6 comms
      use ccsm3_time_management                  ! ccsm3 ocean time
      use ccsm3_global_reductions                ! ccsm3 ocean mpi comms
      use shr_timer_mod                          ! ccsm3 shared timing routines

      implicit none

      integer i,ios,j,k,l,m,n     ,k1m,mm,nn !,mo
      real*8  d1,d2,d3,d4,ddsurf,ddiagf,drstrf,
     &        dske,dskea,dsms,dsmsa,dsmt,dsmta,dsum,dsuma,
     &        dtime,dtime0,dbimon,dmonth,dyear,dyear0
      real    smr,sms,smt,sum,smin,smax,
     &        coord,day2,x,x1,time0,timav,cold,utotp,vtotp
      integer nod,
     &        lt,ma0,ma1,ma2,ma3,mc0,mc1,mc2,mc3,
     &           mk0,mk1,mk2,mk3,mr0,mr1,mr2,mr3,mnth
      logical diagsv,hisurf,histry,restrt
      character intvl*3

      character flnm*80
      character flnmh*80
      character ccsm3_string*80
c
      real*8     dsmall,dsmall2,days1,days6
      parameter (dsmall=0.0001d0,dsmall2=0.0002d0)
      parameter (days1=1.d0,days6=6.d0)
c
      character  charac(50)*1
      integer    nflip
c
      data charac/'1','2','3','4','5','6','7','8','9','0',
     &            '1','2','3','4','5','6','7','8','9','0',
     &            '1','2','3','4','5','6','7','8','9','0',
     &            '1','2','3','4','5','6','7','8','9','0',
     &            '1','2','3','4','5','6','7','8','9','0'/
      data nflip/0/
c
      include 'stmt_fns.h'

 
      integer termination     !if 1, stop now!
      logical end_month       !, ice_ts
      integer ix0,ix1,ix2
      logical bcheck          !check (write) model 'progress' 
      data bcheck /.false./    !if model goes ok, switch it to .false.


!     initial check:-----------------------------------------------
      real, dimension(itdm,jtdm) :: bchk
      real, dimension(imt1:imt2,jmt1:jmt2) :: uu1
!------------------------------------------------------------------ 
       
      lp = 6

!--------------------------------------------------------
!     intialize memory-statistics code
!--------------------------------------------------------
!     call start()
       
!--------------------------------------------------------
!     intialize message-passing with ccsm3 coupler cpl6 
!--------------------------------------------------------
      call ccsm3_setup_coupling_env()

!--------------------------------------------------------
!     intialize i/o for ccsm3 coupled system
!--------------------------------------------------------
      call ccsm3_io_init
 
c --- initialize SPMD processsing
       
      call xcspmd
       
      if (mnproc.eq.1 .and. bcheck) then
        write(lp,*)'(hycom) calling xcspmd done!'
        call flush(lp)
      endif

c
c --- initialize timer names.
c
      call xctmrn(40,'cnuity')
      call xctmrn(41,'tsadvc')
      call xctmrn(42,'momtum')
      call xctmrn(43,'barotp')
      call xctmrn(44,'thermf')
      call xctmrn(45,'ic****')
      call xctmrn(46,'mx****')
      call xctmrn(47,'conv**')
      call xctmrn(48,'diapf*')
      call xctmrn(49,'hybgen')
      call xctmrn(50,'trcupd')
      call xctmrn(51,'restrt')
      call xctmrn(52,'overtn')
      call xctmrn(53,'archiv')

c --- machine-specific initialization
      call machine
 
c --- initialize array i/o.
      call zaiost
 
c --- initiate named-pipe comparison utility
      call pipe_init
 
c --- initialize common variables
 
      call blkdat

      if(mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling blkdat done'
        write(lp,*) '(hycom) thermo = ', thermo
        write(lp,*) '(hycom) hybrid = ', hybrid
        write(lp,*) '(hycom) isopyc = ', isopyc
        call flush(lp)
      endif

      if     (dsurfq.ge.1.0) then
        ddsurf = dsurfq
      elseif (dsurfq.ne.0.0) then
        ddsurf = (baclin/86400.0d0)*
     &           max(1,nint((86400.0d0*dsurfq)/baclin))
      else !no suurface archives
        ddsurf = 999999.0d0*(baclin/86400.0d0)
      endif
      if     (diagfq.ge.1.0) then
        ddiagf = diagfq
      elseif (diagfq.ne.0.0) then
        ddiagf = (baclin/86400.0d0)*
     &           max(1,nint((86400.0d0*diagfq)/baclin))
      else !no 3-d archives
        ddiagf = huge*(baclin/86400.0d0)
      endif
      if     (rstrfq.eq.0.0) then  ! no restart
        drstrf = rstrfq
      elseif (rstrfq.lt.0.0) then  ! no restart at end of run
        drstrf = -rstrfq
      elseif (rstrfq.ge.1.0) then
        drstrf = rstrfq
      else
        drstrf = (baclin/86400.0d0)*
     &           max(1,nint((86400.0d0*rstrfq)/baclin))
      endif
      if (mnproc.eq.1 .and. bcheck) then
      write(lp,*)
      write(lp,*) 'ddsurf = ',ddsurf,nint((86400.0d0*ddsurf)/baclin)
      write(lp,*) 'ddiagf = ',ddiagf,nint((86400.0d0*ddiagf)/baclin)
      write(lp,*) 'drstrf = ',drstrf,nint((86400.0d0*drstrf)/baclin)
      write(lp,*)
      write (lp,101) thkdf2,temdf2,
     &               thkdf4,
     &               veldf2,visco2,
     &               veldf4,visco4,
     &               diapyc,vertmx
 101  format (
     &  ' turb. flux parameters:',1p/
     &  ' thkdf2,temdf2 =',2e9.2/
     &  ' thkdf4        =',1e9.2/
     &  ' veldf2,visco2 =',2e9.2/
     &  ' veldf4,visco4 =',2e9.2/
     &  ' diapyc,vertmx =',2e9.2/)
      endif !1st tile

c
c --- 'lstep' = number of barotropic time steps per baroclinic time step.
c --- lstep   m u s t   be even.
c
      lstep=baclin/batrop
      lstep=2*((lstep+1)/2)
      dlt=baclin/lstep
      if (mnproc.eq.1) then
        write (lp,'(i4,a/)')
     &  lstep,' barotropic steps per baroclinic time step'
        call flush(lp)
      endif !1st tile

!     number of baroclinic time steps per day...
      nsteps_per_day = 86400.0/baclin
c
c --- set up parameters defining the geographic environment
c
        
      call geopar

!     initial checking purpose--are the masks all right?
      IF (bcheck) THEN 
      master_task = 0
      uu1 = float(ip)
      call xcaget(bchk,uu1,master_task)
      if (mnproc == 1) then
        write(lp,*) 'P mask got in main after geopar (1-90): '
        do j=jtdm,1,-1
          write(lp,'(90i1)')(int(bchk(i,j)),i=1,itdm/2)
        enddo
        write(lp,*) 'P mask got in main after geopar (91-180): '
        do j=jtdm,1,-1
          write(lp,'(90i1)')(int(bchk(i,j)),i=itdm/2+1,itdm)
        enddo
        call flush(lp)
      endif
      uu1 = float(iu)
      call xcaget(bchk,uu1,master_task)
      if (mnproc == 1) then
        write(102,*) 'U mask got in main after geopar (1-90): '
        do j=jtdm,1,-1
          write(102,'(90i1)')(int(bchk(i,j)),i=1,itdm/2)
        enddo
        write(102,*) 'U mask got in main after geopar (91-180): '
        do j=jtdm,1,-1
          write(102,'(90i1)')(int(bchk(i,j)),i=itdm/2+1,itdm)
        enddo
        call flush(102)
      endif
      uu1 = float(iv)
      call xcaget(bchk,uu1,master_task)
      if (mnproc == 1) then
        write(103,*) 'V mask got in main after geopar (1-90): '
        do j=jtdm,1,-1
          write(103,'(90i1)')(int(bchk(i,j)),i=1,itdm/2)
        enddo
        write(103,*) 'V mask got in main after geopar (91-180): '
        do j=jtdm,1,-1
          write(103,'(90i1)')(int(bchk(i,j)),i=itdm/2+1,itdm)
        enddo
        call flush(103)
      endif
      ENDIF

      if (mnproc.eq.1 .and. bcheck) then
        write(lp,*)'(hycom) calling geopar done '
        call flush(lp)
      endif !1st tile

 
c --- set up forcing functions
 
!----------------------------------------------------------------------
!     calls to forfuna, forfunk, and forfunp have been removed, because
!     in the coupled version, forcing fields are received from the coupler
!     at the beginning of each coupled interval
!----------------------------------------------------------------------

!     forfunr opens relax fields, must be called here in any case.
      call forfunr  ! bimonthly/monthly climatology

      watcum=0.
      empcum=0.
c
c --- set minimum salinity for each isopycnic layer
      do k=2,kk
!!!!!!!!!!!!!!!!!!!!
        cold=-1.8
!!!     cold=-3.0
!!!!!!!!!!!!!!!!!!!!
        salmin(k)=(sigma(k)-c1-cold*(c2+cold*(c4+c6*cold)))/
     &      (c3+cold*(c5+c7*cold))
      enddo

 
c --- layer specific volume is defined as (1-theta)*thref
c --- subtract constant 'thbase' from theta to reduce roundoff errors
 
      do k=1,kk
        theta(k)=sigma(k)-thbase
      enddo
 
!--------------------------------------------------------------
!     initialize time-related variables 
!--------------------------------------------------------------
      call ccsm3_time_init (time0, time, dtime0, dtime)

 
      if     (mxlkpp) then
c ---   initialize kpp mixing
        call inikpp
      elseif (mxlmy) then
c ---   initialize m-y 2.5 mixing
        call inimy
      elseif (mxlgiss) then
c ---   initialize nasa giss mixing
        call inigiss
      endif
c
      if (linit) then
        delt1 = baclin
        dhdx  = c0p
        dhdy  = c0p
        QICE  = c0p
        AQICE = c0p
        QFLUX = c0p
c
c ---   set up initial conditions
c
        mnth = imonth
Cpg
      if (mnproc.eq.1 .and. bcheck) then
        write (lp,*) '(hycom) Calling inicon, with mnth= ', mnth
        call flush(lp)
      endif 
Cpg
        call inicon(mnth)
Cpg
      if (mnproc.eq.1 .and. bcheck) then
        write (lp,*)'(hycom) calling inicon done '
        call flush(lp)
      endif 
Cpg


        trcrin = .false.
        call initrc(mnth)

        if (mnproc.eq.1 .and. bcheck) then
          write(lp,*)'(hycom) calling inicon and initrc done'
          call flush(lp)
        endif !1st tile

c
c ---   output to archive file
c
!-----------------------------------------------------
!      in coupled version, archiving is not done here
!      DBI blocked off code; NJN removed it
!-----------------------------------------------------
c
      else                   !continue run
        !
        delt1=baclin+baclin
        !
c
c ---   start from restart file
c
        !
        if (mnproc == 1) then
        open(1,file=trim(pointer_filename),form='formatted',status='old')
        read(1,'(a)') flnm
        read(1,'(a)') flnmh
        read(1,*)time                        ! real,  in days
        read(1,*)dtime                       ! real8, in days
        read(1,*)nstep,iyear,imonth,iday     ! integer
     &,          elapsed_days
        read(1,*)eom, eoy                    ! logical
        close(1)
        write(lp,*) ' flnm  = ', flnm
        write(lp,*) ' flnmh = ', flnmh
        write(lp,*) ' pointer_filename = ', pointer_filename
        call flush (lp)
        write(lp,'(2a)')'(hycom) the restart info from: ',trim(pointer_filename)
        write(lp,'(2a)')'     restart file= ',trim(flnm)
        write(lp,*)     '            time= ',time
        write(lp,*)     '           nstep= ',nstep
        write(lp,*)     '           iyear= ',iyear
        write(lp,*)     '          imonth= ',imonth
        write(lp,*)     '            iday= ',iday
        write(lp,*)     '    elapsed_days= ',elapsed_days
        write(lp,*)     '             eom= ', eom
        write(lp,*)     '             eoy= ', eoy
        call flush(lp)
        endif
        !broadcast the above info to all processors:
        call broadcast_scalar(flnm,         master_task)
        call broadcast_scalar(flnmh,        master_task)
        call broadcast_scalar(time,         master_task)
        call broadcast_scalar(dtime,        master_task)
        call broadcast_scalar(nstep,        master_task)
        call broadcast_scalar(iyear,        master_task)
        call broadcast_scalar(imonth,       master_task)
        call broadcast_scalar(iday,         master_task)
        call broadcast_scalar(elapsed_days, master_task)
        call broadcast_scalar(eom,          master_task)
        call broadcast_scalar(eoy,          master_task)

    
!-----------------------------------------------------------------------
!     make any other needed adjustments after header file has been read
!     and report restart time in format recognized by ccsm3
!-----------------------------------------------------------------------
        call ccsm3_time_init1

        call restart_in(nstep,dtime, flnm, flnmh)

        if (mnproc == 1 ) then ! .and. bcheck) then
            write(lp,*)'(hycom) restartfile reading successful!'
        endif
        !
         time0 = (iyear0-1)*days_in_year+days_in_prior_months(imonth0)
     &                                  +iday0+0.0001
        !the above makes time0 be the end of last integration (in days)!
        !
        !initialize annual and monthly sum here ???
       
      endif !initial conditions

c
c --- set barotp.pot.vort. and layer thickness (incl.bottom pressure) at
c --- u,v points
c
      call dpthuv

      if (mnproc == 1 .and. bcheck) then
        write(lp,*) '(hycom) calling dpthuv done'
        call flush(lp)
      endif
 
!!    call xctilr(dp(    1-nbdy,1-nbdy,1,1),1,2*kk, nbdy,nbdy, halo_ps)
!--------------------------- begin nonrecursive xctilr ---------------
      call xctilr(dp(    1-nbdy,1-nbdy,1,1),1,kk, nbdy,nbdy, halo_ps)
      call xctilr(dp(    1-nbdy,1-nbdy,1,2),1,kk, nbdy,nbdy, halo_ps)
!------------------------end nonrecursive xctilr ----------------------

      call xctilr(dpmixl(1-nbdy,1-nbdy,  1),1,   2, nbdy,nbdy, halo_ps)
      call xctilr(thkk(  1-nbdy,1-nbdy    ),1,   1, nbdy,nbdy, halo_ps)
      call xctilr(psikk( 1-nbdy,1-nbdy    ),1,   1, nbdy,nbdy, halo_ps)
 
      margin = nbdy

      do n=1,2
      m=mod(n,2)+1
 
!$OMP   PARALLEL DO PRIVATE(j,k,l,i)
!$OMP&           SCHEDULE(STATIC,jblk)
        do j=1-margin,jj+margin
          do l=1,isp(j)
            if     (n.eq.mod(nstep+1,2)+1) then
              do i=max(1-margin,ifp(j,l)),min(ii+margin,ilp(j,l))
                dpbl(i,j)=dpmixl(i,j,n)
              enddo
            endif
            do i=max(1-margin,ifp(j,l)),min(ii+margin,ilp(j,l))
              p(i,j,1)=0.0
              do k=1,kk
                p(i,j,k+1)=p(i,j,k)+dp(i,j,k,n)
              enddo
            enddo
          enddo
        enddo
!$OMP   END PARALLEL DO
c
        call dpudpv(dpu(1-nbdy,1-nbdy,1,n),
     &              dpv(1-nbdy,1-nbdy,1,n),
     &              p,depthu,depthv, max(0,margin-1))

        if (mnproc == 1 .and. bcheck) then
          write(lp,*) '(hycom) calling dpudpv done at n= ',n
          call flush(lp)
        endif
 
 
      enddo  !n=1,2
 
 
      timav=time0
      m=mod(nstep  ,2)+1
      n=mod(nstep+1,2)+1

           call pipe_comparall(m,n, 'restrt, step')
 

!----------------------------------------------------------
!     in the coupled model, forcing is received from the
!     coupler; therefore code for reading stand-alone version
!     forcing fields has been removed
!----------------------------------------------------------

      m=mod(nstep  ,2)+1
      n=mod(nstep+1,2)+1
     
      k1m=m                !NOT k1m=1+mm
      k1n=n                !NOT k1n=1+nn
      

      !-------------------------------------------------------------
      ! initialize namelists, grid information, and initial message-
      ! passing for cpl6
      !-------------------------------------------------------------
      call ccsm3_init_coupled(flnmgrdd)

      if (mnproc == 1) then
        write(lp,*)'(hycom) ccsm3_init_coupled completed'
        call flush(lp)
      endif


      if (mnproc == 1) then
        write (*,*) '(hycom) model starts from day, nstep=', iday,nstep
      endif

c
c --- report initialization time.
c
      call xctmrp


!-----------------------------------------------------------------
!     eliminate code which reads relaxation climatology fields
!     (not used when coupled)
!-----------------------------------------------------------------


      if     (bnstfq.ne.0.0) then  ! initialize barotropic boundary input
        wb0=-99.0
        wb1=-99.0
        call rdbaro(dtime0)
      endif
 
      if     (nestfq.ne.0.0) then  ! initialise 3-d nesting input
        wn0=-99.0
        wn1=-99.0
        call rdnest(dtime0)
      endif

      nod = 14
      write(cyear,'(i4.4)') iyear
Cpg - Write out summary files for each month
      write(flnm,'(a,i4.4,a,i2.2)')'summary_out.',iyear,'-',imonth    
      flnmsum=trim(flnmsumd)//trim(flnm)
      if (mnproc==1) then
        open(unit=nod,file=flnmsum,status='unknown')
        write(nod,'(3a,i2.2)')
     &    'Summary of diagnostic outputs for year ',cyear
     &  , '  month ', imonth
Cpg
        write(nod,*)
        call flush(nod)
      endif
c
c **************************************************************************
c
c ************************* main loop starts here **************************
c
c **************************************************************************
c
c --- letter 'm' refers to mid-time level (example: dp(i,j,k,m) )
c --- letter 'n' refers to old and new time level
c


!-----------------------------------------------------------------------
!     note that in ccsm3, the coupler controls when each component model
!     stops by sending a "stop now" flag.  The following loop is therefore
!     defined to be an infinite loop; the loop is exited when the cpl
!     sends a stop signal (stop_now)
!-----------------------------------------------------------------------

!     if (mnproc == 1) then
        write(lp,*) ' '
        write(lp,*) ' =================================================='
        write(lp,*) ' '
        write(lp,*)'(hycom) Begin main loop -- receive forcing from cpl6'
        write(lp,*) ' '
        call flush(lp)
!     endif

       
      MAIN: do 

      m=mod(nstep  ,2)+1
      n=mod(nstep+1,2)+1

      k1m=m              !NOT k1m=1+mm (DBI)
      k1n=n              !NOT k1n=1+nn (DBI)
      

      if (mnproc == 1 .and. nsteps_run .gt. 0 .and. bcheck) then
        write(lp,*)' end of timestep ', nstep
        write(lp,*) ' '
        call flush(lp)
      endif

!-------------------------------------------------------------------------
!     exchange fields and fluxes with the ccsm3 flux coupler
!-------------------------------------------------------------------------
      call ccsm3_set_coupled_forcing

!--------------------------------------------------------
!     stop if cpl has sent the stop signal
!--------------------------------------------------------
      if (stop_now == 1) exit MAIN


!-------------------------------------------------------------------------
!     ccsm3-related time-keeping
!-------------------------------------------------------------------------
      call ccsm3_time_advance (time, dtime)


!-------------------------------------------------------------------------
!     DBI identified the blocks with forcing weights as unnecessary in 
!     the coupled version; NJN removed the code (1/3/2005)
!-------------------------------------------------------------------------

      IF (.true.) THEN
      if     (bnstfq.ne.0.0) then  ! wieghts for baro nesting
        call rdbaro(dtime)
      endif
 
      if     (nestfq.ne.0.0) then  ! wieghts for 3-d  nesting
        call rdnest(dtime)
      endif
      ENDIF


         call pipe_comparall(m,n, 'ENTER , step')
         call xctmr0(40)
      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling cnuity...'
        call flush(lp)
      endif

      call cnuity(m,n)
         call xctmr1(40)
         call pipe_comparall(m,n, 'cnuity, step')
         call xctmr0(41)
      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling tsadvc...'
        call flush(lp)
      endif


      call tsadvc(m,n)
         call xctmr1(41)
         call pipe_comparall(m,n, 'tsadvc, step')
         call xctmr0(42)


      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling momtum...'
        call flush(lp)
      endif
      call momtum(m,n)
      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling momtum done'
        call flush(lp)
      endif
         call xctmr1(42)
         call pipe_comparall(m,n, 'momtum, step')
         call xctmr0(43)


      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling barotp...'
        call flush(lp)
      endif
      call barotp(m,n)
         call xctmr1(43)
         call pipe_comparall(m,n, 'barotp, step')
         call xctmr0(44)
      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling thermf...'
        call flush(lp)
      endif


      IF (.false.) THEN
      call thermf(m,n)
         call xctmr1(44)
         call pipe_comparall(m,n, 'thermf, step')
      if (icegln) then
            call xctmr0(45)
        if (mnproc == 1 .and. bcheck) then
         write(lp,*)'(hycom) calling icloan...'
         call flush(lp)
        endif
         call icloan(m,n)
            call xctmr1(45)
            call pipe_comparall(m,n, 'icloan, step')
      else
            call xctmr0(45)
            call xctmr1(45)
      endif !icegln
      ENDIF
!     checking:
      IF (.false. .and. mod(nstep,nstep0)==1 ) THEN
        call xcaget(bchk,salflx,0)
        if (mnproc == 1) then
          write(102,'(10e13.5)')bchk
          call flush(102)
        endif
      ENDIF
!----------
!     thermf_c is largely modified (from thermf) for coupling case!
      !---
      !!!covice = aice        
      !ice concentration covice is now obtained in recv_from_cp directively.
      !
      
      call thermf_c(m,n)   !,mm,nn,k1m,k1n)
      !---
      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) calling thermf_c done.' 
        call flush(lp)
      endif
!     checking:
      IF (.false. .and. mod(nstep,nstep0)==1 ) THEN
        call xcaget(bchk,salflx,0)
        if (mnproc == 1) then
          write(103,'(10e13.5)')bchk
          call flush(103)
        endif
      ENDIF
!-----------
      !get potential frezing/melting heat flux needed for ice model.
      !mixed layer T and S are adjusted only TWICE in each coupling interval.
      !

      call ice_formation(k1n, ice_ts)


      if (mnproc == 1 .and. bcheck .and. ice_ts) then
        write(lp,*)'(hycom) called ice_formation at nstep=',nstep
        call flush(lp)
      endif
!DBI: checking:
!      IF (.true. .and. mod(nstep,nstep0)==1 ) THEN
!        call xcaget(bchk,salflx,0)
!        if (mnproc == 1) then
!          write(104,'(10e13.5)')bchk
!          call flush(104)
!        endif
!      ENDIF
!-----------

      !===================================================================
      !if (.false. .and. ice_ts) then   !DBI: switched this call off!
      !  call ice_flx_to_coupler        ! work done in ice_formation now!    
      !  if (mnproc == 1 .and. bcheck) then
      !  write(lp,*)'DBI: (mainloop) called ice_flx_to_coupler at nstep=',nstep
      !  call flush(lp)
      !  endif
      !endif
!----
      if (trcout) then
           call xctmr0(50)
        call trcupd(m,n)
           call xctmr1(50)
           call pipe_comparall(m,n, 'trcupd, step')
      endif !trcout


      if (hybrid) then
         diagsv = diagno
         diagno = diagno .or. nstep.eq.nstep0+1 .or.
     &            histry .or. hisurf .or.
     &            mod(dtime+dsmall,days6).lt.dsmall2
     &            .or.mod(nstep,nsteps_per_day/24).eq.0  !hrly
         if (mxlkpp .or. mxlmy .or. mxlgiss) then
           if (mnproc == 1 .and. bcheck) then
             write(lp,*)'(hycom) calling mxkprf...'
             call flush(lp)
           endif
               call xctmr0(46)
            call mxkprf(m,n)
           if (mnproc == 1 .and. bcheck) then
             write(lp,*)'(hycom) calling mxkprf...'
             call flush(lp)
           endif
               call xctmr1(46)
               call pipe_comparall(m,n, 'mxkprf, step')
         elseif (mxlpwp) then
           if (mnproc == 1 .and. bcheck) then
             write(lp,*)'(hycom) calling mxpwp...'
             call flush(lp)
           endif
               call xctmr0(46)
            call mxpwp(m,n)
               call xctmr1(46)
               call pipe_comparall(m,n, 'mxpwp,  step')
         elseif (mxlkta) then
           if (mnproc == 1 .and. bcheck) then
             write(lp,*)'(hycom) calling mxkrta...'
             call flush(lp)
           endif
               call xctmr0(46)
            call mxkrta(m,n)
               call xctmr1(46)
               call pipe_comparall(m,n, 'mxkrta, step')
         elseif (mxlktb) then
           if (mnproc == 1 .and. bcheck) then
             write(lp,*)'(hycom) calling mxkrtb...'
             call flush(lp)
           endif
               call xctmr0(46)
            call mxkrtb(m,n)
               call xctmr1(46)
               call pipe_comparall(m,n, 'mxkrtb, step')
         else
               call xctmr0(46)
               call xctmr1(46)
         endif
         diagno = diagsv
         if (mxlkpp .or. mxlmy .or. mxlgiss) then
               call xctmr0(47)
               call xctmr1(47)
               call xctmr0(48)
               call xctmr1(48)
         else  ! mxlpwp has dypflg=2
           if (mnproc == 1 .and. bcheck) then
             write(lp,*)'(hycom) calling convch...'
             call flush(lp)
           endif
               call xctmr0(47)
            call convch(m,n)
               call xctmr1(47)
               call pipe_comparall(m,n, 'convch, step')
           if     (dypflg.eq.1) then  ! KPP-like
             if (mnproc == 1 .and. bcheck) then
               write(lp,*)'(hycom) calling diapf1...'
               call flush(lp)
             endif
                 call xctmr0(48)
              call diapf1(m,n)
                 call xctmr1(48)
                 call pipe_comparall(m,n, 'diapf1, step')
           elseif (dypflg.eq.2) then  ! explicit
             if (mnproc == 1 .and. bcheck) then
               write(lp,*)'(hycom) calling diapf2...'
               call flush(lp)
             endif
                 call xctmr0(48)
              call diapf2(m,n)
                 call xctmr1(48)
                 call pipe_comparall(m,n, 'diapf2, step')
           else
                 call xctmr0(48)
                 call xctmr1(48)
           endif
        endif

          if (mnproc == 1 .and. bcheck) then
            write(lp,*)'(hycom) calling hybgen...'
            call flush(lp)
          endif
           call xctmr0(49)
        call hybgen(m,n)
           call xctmr1(49)
           call pipe_comparall(m,n, 'hybgen, step')
      else  ! isopyc
            call xctmr0(46)
         call mxkrtm(m,n)
            call xctmr1(46)
            call pipe_comparall(m,n, 'mxkrtm, step')
            call xctmr0(47)
         call convcm(m,n)
            call xctmr1(47)
            call pipe_comparall(m,n, 'convcm, step')
            call xctmr0(48)
         call diapf3(m,n)
            call xctmr1(48)
            call pipe_comparall(m,n, 'diapf3, step')
            call xctmr0(49)
            call xctmr1(49)
      endif !hybrid:isopyc


      call ssh(m)               !get sea surface height gradient
      if (mnproc == 1 .and. bcheck) then
        write(lp,*)'(hycom) after called ssh at nstep=',nstep
        call flush(lp)
      endif


c ---------------------------------------------------------------------------
c
c --- output and diagnostic calculations
c
c ---------------------------------------------------------------------------
c
      if (diagno .or. histry .or. hisurf .or.
     &    nstep.eq.nstep0+1 .or.
     &    mod(dtime+dsmall,days1).lt.dsmall2) then  ! at least daily
c
!----
c
c ---   diagnose mean sea surface height


        smin= huge
        smax=-huge
!$OMP   PARALLEL DO PRIVATE(j,l,i)
!$OMP&              REDUCTION(MIN:smin) REDUCTION(MAX:smax)
!$OMP&           SCHEDULE(STATIC,jblk)
        do j=1,jj
          do l=1,isp(j)
            do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
c ---         compute sea surface height
              util1(i,j)=montg(i,j,1)+thref*pbavg(i,j,m)
              util3(i,j)=util1(i,j)  *scp2(i,j)
              util4(i,j)=montg(i,j,1)*scp2(i,j)
              smin=min(smin,util1(i,j))
              smax=max(smax,util1(i,j))
            enddo
          enddo
        enddo
!$OMP   END PARALLEL DO
        call xcminr(smin)
        call xcmaxr(smax)
        call xcsum( dsum, util3,ip)
        call xcsum( dsmt, util4,ip)
        sum=dsum
        smt=dsmt

        if (mnproc.eq.1) then
        write (lp,'(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &              '' mean sea srf.hgt. (mm):'',f8.2,
     &              ''  ('',1pe8.1,'' to '',e8.1,'')'')')
     &  nstep,iyear,imonth,iday,ihour,
     &  sum/(area*thref*onemm),smin/(thref*onemm),smax/(thref*onemm)
*       write (lp,'(i9,'' ('',i4,''/'',i3.3,'')'',
*    .              '' mean Montg. pot.  (mm):'',f8.2)')
*    .  nstep,iyear,iday,
*    .  smt/(area*thref*onemm)
        call flush(lp)
        write(nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2,'')'',
     &              '' mean sea srf.hgt. (mm):'',f8.2,
     &              ''  ('',1pe8.1,'' to '',e8.1,'')'')')
     &  nstep,iyear,imonth,iday,ihour,
     &  sum/(area*thref*onemm),smin/(thref*onemm),smax/(thref*onemm)
*       write(nod,'(i9,'' ('',i4,''/'',i3.3,'')'',
*    .              '' mean Montg. pot.  (mm):'',f8.2)')
*    .  nstep,iyear,iday,
*    .  smt/(area*thref*onemm)
        call flush(nod)
        endif !1st tile




c ---   NaN detection.
        if     (smin.ne.smin .or. sum.ne.sum .or.
     &          smax.ne.smax .or. smt.ne.smt      ) then

          if (mnproc.eq.1) then
          write(lp,*)
          write(lp,*) 'error - NaN detected'
          write(lp,*)
          call flush(lp)
          endif !1st tile
          call ccsm3_exit_HYCOM('ERROR: NaN detected',normal_exit=.false.)
         
        endif
      endif
c
c --- diagnose heat/salt flux, ice, layer thickness and temperature,
c --- mean temperature and mean kinetic energy
c --- note that mixed-layer fields must be switched on in mxkprf
      if (diagno .or.
     &    nstep.eq.nstep0+1 .or.
     &    mod(nstep,nsteps_per_day/24).eq.0 )  then !hrly
!     &    mod(dtime+dsmall,days6).lt.dsmall2) then  ! at least every 6 days
!...check everyday:
!     &    mod(dtime+dsmall,days1).lt.dsmall2) then  ! at least daily
c
!----
c
        if (thermo) then
!$OMP     PARALLEL DO PRIVATE(j,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                util1(i,j)=buoflx(i,j)*scp2(i,j)
                util2(i,j)=bhtflx(i,j)*scp2(i,j)
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
          call xcsum(dsum, util1,ip)
          call xcsum(dsmt, util2,ip)
          sum= (dsum*1.00D9)/area  ! 1.e9*m**2/sec**3
          smt= (dsmt*1.00D9)/area  ! 1.e9*m**2/sec**3
          if (mnproc.eq.1) then
          write (lp, '(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &        '' mean B. flux (m^2/s^3):'',f8.2,
     &                           '' hfl:'',f8.2)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt
          call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &        '' mean B. flux (m^2/s^3):'',f8.2,
     &                           '' hfl:'',f8.2)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt
          call flush(nod)
          endif !1st tile
c
!$OMP     PARALLEL DO PRIVATE(j,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                util1(i,j)=surflx(i,j)*scp2(i,j)
                util2(i,j)=mixflx(i,j)*scp2(i,j)
                util3(i,j)=sswflx(i,j)*scp2(i,j)
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
          call xcsum(dsum, util1,ip)
          call xcsum(dsmt, util2,ip)
          call xcsum(d3,   util3,ip)
          sum= dsum/area
          smt= dsmt/area
          smr= d3  /area
          if (mnproc.eq.1) then
          write (lp, '(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &        '' mean heat flux (w/m^2):'',f8.2,
     &                           '' ssw:'',f8.2,
     &                           ''  ml:'',f8.2)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smr,smt
          call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &        '' mean heat flux (w/m^2):'',f8.2,
     &                           '' ssw:'',f8.2,
     &                           ''  ml:'',f8.2)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smr,smt
          call flush(nod)
          endif !1st tile
!      check surflx, sswflx, salflx and sssflx:
          if (.false.) then
            write (lp,*)'(hycom) checking fluxes at nstep=',nstep
            call flush(lp) 
            call xcaget(bchk,surflx,0)
            open(22,file='dat_surflx.main',form='formatted',
     .              position='append') 
            write(22,'(10e13.5)')bchk
            call flush(22)
            close(22)
            call xcaget(bchk,sswflx,0)
            open(22,file='dat_sswflx.main',form='formatted',
     .              position='append')
            write(22,'(10e13.5)')bchk
            call flush(22)
            close(22)
            call xcaget(bchk,sstflx,0)
            open(22,file='dat_sstflx.main',form='formatted',
     .              position='append')
            write(22,'(10e13.5)')bchk
            call flush(22)
            close(22)
            call xcaget(bchk,salflx,0)
            open(22,file='dat_salflx.main',form='formatted',
     .              position='append')
            write(22,'(10e13.5)')bchk
            call flush(22)
            close(22)
            call xcaget(bchk,sssflx,0)
            open(22,file='dat_sssflx.main',form='formatted',
     .              position='append')
            write(22,'(10e13.5)')bchk
            call flush(22)
            close(22)
          endif
!----------------------------
c
!$OMP     PARALLEL DO PRIVATE(j,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                util1(i,j)=salflx(i,j)*scp2(i,j)/saln(i,j,1,n)
                util2(i,j)=sssflx(i,j)*scp2(i,j)/saln(i,j,1,n)
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
          call xcsum(dsms, util1,ip)
          call xcsum(dsum, util2,ip)
          sms=-(dsms*thref*7.0D0*8.64D7)/area  ! P-E in mm/week
          smr=-(dsum*thref*7.0D0*8.64D7)/area  ! P-E in mm/week
          if (mnproc.eq.1) then
          write (lp, '(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &        '' mean P-E  flux (mm/wk):'',f8.2,
     &                           '' sss:'',f8.2)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sms,smr
          call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &        '' mean P-E  flux (mm/wk):'',f8.2,
     &                           '' sss:'',f8.2)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sms,smr
          call flush(nod)
          endif !1st tile
        endif
c
        if (icegln) then  ! basin-wide ice
!$OMP     PARALLEL DO PRIVATE(j,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                if     (covice(i,j).ne.0.0) then
                  util1(i,j)=                        scp2(i,j)
                  util2(i,j)=covice(i,j)*            scp2(i,j)
                  util3(i,j)=            thkice(i,j)*scp2(i,j)
                  util4(i,j)=covice(i,j)*temice(i,j)*scp2(i,j)
                else
                  util1(i,j)=0.0
                  util2(i,j)=0.0
                  util3(i,j)=0.0
                  util4(i,j)=0.0
                endif
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
          call xcsum(d1, util1,ip)
          call xcsum(d2, util2,ip)
          call xcsum(d3, util3,ip)
          call xcsum(d4, util4,ip)
          if     (d2.gt.0.0d0) then
            sum=d3/d1
            smt=d4/d2
            sms=d2/area * 100.0
          else
            sum=0.0 
            smt=0.0 
            sms=0.0 
          endif
          if (mnproc.eq.1) then
          write (lp,'(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &              '' mean      ice thk. (m):'',f8.2,
     &                               ''  temp:'',f6.2,
     &                               ''  pcen:'',f7.3)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt,sms
          call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &              '' mean      ice thk. (m):'',f8.2,
     &                               ''  temp:'',f6.2,
     &                               ''  pcen:'',f7.3)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt,sms
          call flush(nod)
          endif !1st tile
        endif  ! icegln
c
        if (nreg.ne.0 .and. icegln) then  ! southern hemisphere ice
!$OMP     PARALLEL DO PRIVATE(j,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                if     (covice(i,j).ne.0.0 .and.
     &                    plat(i,j).lt.0.0      ) then
                  util1(i,j)=                        scp2(i,j)
                  util2(i,j)=covice(i,j)*            scp2(i,j)
                  util3(i,j)=            thkice(i,j)*scp2(i,j)
                  util4(i,j)=covice(i,j)*temice(i,j)*scp2(i,j)
                else
                  util1(i,j)=0.0
                  util2(i,j)=0.0
                  util3(i,j)=0.0
                  util4(i,j)=0.0
                endif
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
          call xcsum(d1, util1,ip)
          call xcsum(d2, util2,ip)
          call xcsum(d3, util3,ip)
          call xcsum(d4, util4,ip)
          if     (d2.gt.0.0d0) then
            sum=d3/d1
            smt=d4/d2
            sms=d2/area * 100.0
          else
            sum=0.0 
            smt=0.0 
            sms=0.0 
          endif
          if (mnproc.eq.1) then
          write (lp,'(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &              '' mean S.H. ice thk. (m):'',f8.2,
     &                               ''  temp:'',f6.2,
     &                               ''  pcen:'',f7.3)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt,sms
          call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &              '' mean S.H. ice thk. (m):'',f8.2,
     &                               ''  temp:'',f6.2,
     &                               ''  pcen:'',f7.3)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt,sms
          call flush(nod)
          endif !1st tile
        endif  ! icegln .and. nreg.ne.0
c
!$OMP   PARALLEL DO PRIVATE(j,l,i)
!$OMP&           SCHEDULE(STATIC,jblk)
        do j=1,jj
          do l=1,isp(j)
            do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
              util1(i,j)=dpmixl(i,j,n)*scp2(i,j)
              util2(i,j)=dpmixl(i,j,n)*scp2(i,j)*tmix(i,j)
              util3(i,j)=dpmixl(i,j,n)*scp2(i,j)*smix(i,j)
            enddo
          enddo
        enddo
!$OMP   END PARALLEL DO
        call xcsum(dsum, util1,ip)
        call xcsum(dsmt, util2,ip)
        call xcsum(dsms, util3,ip)
        if     (dsum.ne.0.0d0) then
          sum=dsum/(area*onem)
          smt=dsmt/dsum
          sms=dsms/dsum
        else
          sum=0.0
          smt=0.0
          sms=0.0
        endif
        if (mnproc.eq.1) then
        write (lp,'(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &              '' mean mixlayer thk. (m):'',f8.2,
     &                               ''  temp:'',f6.2,
     &                               ''  saln:'',f7.3)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt,sms
        call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &              '' mean mixlayer thk. (m):'',f8.2,
     &                               ''  temp:'',f6.2,
     &                               ''  saln:'',f7.3)')
     &      nstep,iyear,imonth,iday,ihour,
     &      sum,smt,sms
        call flush(nod)
        endif !1st tile
c
        if     (relaxf) then
c
c ---     mean surface climatology.
c
!$OMP     PARALLEL DO PRIVATE(j,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                if     (sstflg.le.1) then
                  util1(i,j)=scp2(i,j)*
     &                (twall(i,j,1,lc0)*wc0+twall(i,j,1,lc1)*wc1
     &                +twall(i,j,1,lc2)*wc2+twall(i,j,1,lc3)*wc3)
                else !synoptic observed sst
                  util1(i,j)=scp2(i,j)*
     &                (seatmp(i,j,l0)*w0+seatmp(i,j,l1)*w1
     &                +seatmp(i,j,l2)*w2+seatmp(i,j,l3)*w3)
                endif
                util2(i,j)=scp2(i,j)*
     &                (swall(i,j,1,lc0)*wc0+swall(i,j,1,lc1)*wc1
     &                +swall(i,j,1,lc2)*wc2+swall(i,j,1,lc3)*wc3)
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
          call xcsum(dsmt, util1,ip)
          call xcsum(dsms, util2,ip)
          smt=dsmt/area
          sms=dsms/area
          if (mnproc.eq.1) then
          write (lp,'(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &                '' mean surfclim thk. (m):'',f8.2,
     &                                 ''  temp:'',f6.2,
     &                                 ''  saln:'',f7.3)')
     &        nstep,iyear,imonth,iday,ihour,
     &        thkmin,smt,sms
          call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &                '' mean surfclim thk. (m):'',f8.2,
     &                                 ''  temp:'',f6.2,
     &                                 ''  saln:'',f7.3)')
     &        nstep,iyear,imonth,iday,ihour,
     &        thkmin,smt,sms
          call flush(nod)
          endif !1st tile
        endif
c
        call xctilr(u(    1-nbdy,1-nbdy,1,n),1,kk, 1,1, halo_uv)
        call xctilr(v(    1-nbdy,1-nbdy,1,n),1,kk, 1,1, halo_vv)
        call xctilr(ubavg(1-nbdy,1-nbdy,  n),1, 1, 1,1, halo_uv)
        call xctilr(vbavg(1-nbdy,1-nbdy,  n),1, 1, 1,1, halo_vv)
c
        dsuma=0.0d0
        dsmta=0.0d0
        dsmsa=0.0d0
        dskea=0.0d0
        do k=1,kk
!$OMP     PARALLEL DO PRIVATE(j,l,i,utotp,vtotp)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                utotp=0.5*( u(i,  j,k,n)+ubavg(i,  j,n) +
     &                      u(i+1,j,k,n)+ubavg(i+1,j,n)  )
                vtotp=0.5*( v(i,j,  k,n)+vbavg(i,j,  n) +
     &                      v(i,j+1,k,n)+vbavg(i,j+1,n)  )
                util1(i,j)=dp(i,j,k,n)*scp2(i,j)
                util2(i,j)=dp(i,j,k,n)*scp2(i,j)*temp(i,j,k,n)
                util3(i,j)=dp(i,j,k,n)*scp2(i,j)*saln(i,j,k,n)
                util4(i,j)=dp(i,j,k,n)*scp2(i,j)*
     &                    0.5*(1000.0+th3d(i,j,k,n)+thbase)*
     &                        (utotp**2+vtotp**2)
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO

 

          call xcsum(dsum, util1,ip)
          call xcsum(dsmt, util2,ip)
          call xcsum(dsms, util3,ip)
          call xcsum(dske, util4,ip)
          dsuma=dsuma+dsum
          dsmta=dsmta+dsmt
          dsmsa=dsmsa+dsms
          dskea=dskea+dske
          if     (dsum.ne.0.0d0) then
            sum=dsum/(area*onem)
            smt=dsmt/dsum
            sms=dsms/dsum
          else
            sum=0.0
            smt=0.0
            sms=0.0
          endif
          if (mnproc.eq.1) then
          write (lp,'(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &                '' mean layer '',i2,'' thk. (m):'',f8.2,
     &                                       ''  temp:'',f6.2,
     &                                       ''  saln:'',f7.3)')
     &        nstep,iyear,imonth,iday,ihour,
     &        k,sum,smt,sms
          call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &                '' mean layer '',i2,'' thk. (m):'',f8.2,
     &                                       ''  temp:'',f6.2,
     &                                       ''  saln:'',f7.3)')
     &        nstep,iyear,imonth,iday,ihour,
     &        k,sum,smt,sms
          call flush(nod)
          endif !1st tile
        enddo
        sum=dskea/(area*onem)
        smt=dsmta/dsuma
        sms=dsmsa/dsuma
        if (mnproc.eq.1) then
        write (lp,'(i9,'' ('',i4,''/'',i2,''/'',i2,''/'',i2,'')'',
     &                '' region-wide mean K.E. :'',f8.2,
     &                                 ''  temp:'',f7.3,
     &                                  '' saln:'',f8.4)')
     &      nstep,iyear,imonth,iday,ihour,
     &        sum,smt,sms
        call flush(lp)
          write (nod,'(i9,'' ('',i4,''/'',i2.2,''/'',i2.2,''/'',i2.2,'')'',
     &                '' region-wide mean K.E. :'',f8.2,
     &                                 ''  temp:'',f7.3,
     &                                  '' saln:'',f8.4)')
     &      nstep,iyear,imonth,iday,ihour,
     &        sum,smt,sms
        call flush(nod)
        endif !1st tile
      endif

c
c --- diagnose meridional overturning and heat flux
!
      if (doovtn) then
        !write(flnm,'(a,i4.4,i2.2,i2.2)')'ovrtn.',iyear,imonth,iday
        write(flnm,'(a,i4.4,a,i2.2)')'ovrtn.',iyear,'-',imonth    
             !PG: now one file for each month!
        flnmovr = trim(flnmovrd)//trim(flnm)
        call xctmr0(52)
        !call overtn(dtime,dyear)      !DBI: dtime, dyear critical?
        call overtn(time,float(days_in_year))
        call xctmr1(52)
        if (mnproc == 1 .and. bcheck) then
          write(lp,*)'    (mainloop) called overtn. time= ',time
          write(lp,*)'      iyear,imonth,iday: ',iyear,imonth,iday
          call flush(lp)
        endif
      endif
      !
      IF (.false.) THEN  
      if     (mod(dtime+dsmall,dmonth).lt.dsmall2) then
        call xctmr0(52)
        call overtn(dtime,dyear)
        call xctmr1(52)
      endif
      ENDIF
c
!
      IF (dohist) THEN    

        call xctmr0(53)
c
c ---   output to archive file
c
!----
c
!$OMP   PARALLEL DO PRIVATE(j,k,l,i)
!$OMP&           SCHEDULE(STATIC,jblk)
        do j=1,jj
          do l=1,isp(j)
            do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
              util1(i,j)=montg(i,j,1)+thref*pbavg(i,j,m)
            enddo
          enddo
          if (isopyc .or. mxlkrt) then
            do l=1,isu(j)
              do i=max(1,ifu(j,l)),min(ii,ilu(j,l))
                umix(i,j)=u(i,j,1,n)
              enddo
            enddo
            do l=1,isv(j)
              do i=max(1,ifv(j,l)),min(ii,ilv(j,l))
                vmix(i,j)=v(i,j,1,n)
              enddo
            enddo
          endif
            do k= 1,kk
              do l=1,isp(j)
                do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
c ---             convert diapycnal thickness changes into 
c ---             actual interface fluxes
                  if (k.gt.1) then
                    diaflx(i,j,k)=diaflx(i,j,k)/(2.*onem) +
     &                            diaflx(i,j,k-1)
                  else
                    diaflx(i,j,k)=diaflx(i,j,k)/(2.*onem)
                  endif
                enddo
              enddo
            enddo
        enddo
!$OMP   END PARALLEL DO
c
        if (mnproc.eq.1) then
        write (intvl,'(i3.3)') int(dtime-timav+dsmall)
        endif !1st tile
        !DBI: new name for archive file
        k = imonth-1           !NOT imonth
        l = iyear
        if (k==0) then
          k=12
          l=iyear-1
        endif
        write(flnm,'(a,i4.4,a,i2.2)')'archv.', l, '-', k 
        flnmarc = trim(flnmarcd)//trim(flnm)
        !if (hisurf .and. .not. histry) then
        !  call archiv(n, 1,  iyear,iday,ihour, intvl)
        !else
          call archiv(n, kk, iyear,iday,ihour, intvl)
        !endif

!DBI: mass fluxes are currently output in restart_out (strange...)
!     should do it in archiv .... later!
        write(flnm,'(a,i4.4,a,i2.2)')'flxdp.', l, '-', k
        flnmflx = trim(flnmflxd)//trim(flnm)
!-----------------------------------------------------------------
c
!$OMP     PARALLEL DO PRIVATE(j,k,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do k= 1,kk
              do l=1,isp(j)
                do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                  diaflx(i,j,k)=0.0
                enddo
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
c
          timav=time
        call xctmr1(53)

        ENDIF         !DBI (dohist) intead of histry.or.hisurf
 
      if (dorestart .or. (cpl_write_restart .and. eod) ) then  


        if (mnproc.eq.1) then
          write(lp,1100) iyear, cmonth3(imonth), iday, isecond
          call flush(lp)
        endif !1st tile
1100  format ('(hycom)', 'write restart at ', i4.4, '-', a3, '-',
     &        i2.2,', ', i7, ' sec')

        call xctmr0(51)
c
c ---   output to restart and flux statitics files
c

        if (mxlkpp) then
!$OMP     PARALLEL DO PRIVATE(j,l,i)
!$OMP&             SCHEDULE(STATIC,jblk)
          do j=1,jj
            do l=1,isp(j)
              do i=max(1,ifp(j,l)),min(ii,ilp(j,l))
                dpmixl(i,j,1) = dpbl(i,j)
                dpmixl(i,j,2) = dpbl(i,j)
              enddo
            enddo
          enddo
!$OMP     END PARALLEL DO
        endif


!================================
        !DBI: (modified by njn01 to meet ccsm3 naming requirements)

        call ccsm3_time_date_stamp (ccsm3_string, 'ymds')

        flnmrso  = trim(flnmrsod)//'.'//trim(ccsm3_string)  ! restart file
        flnmrsoh = trim(flnmrsod)//'h.'//trim(ccsm3_string) ! restart "header" file

        open(1,file=trim(pointer_filename),form='formatted',
     .          status='unknown')
        write(1,'(a)')trim(flnmrso)
        write(1,'(a)')trim(flnmrsoh)
        write(1,*)time                        ! real,  in days
        write(1,*)dtime                       ! real8, in days
        write(1,*)nstep,iyear,imonth,iday     ! integer
     &,          elapsed_days
        write(1,*)eom, eoy                    ! logical
        close(1)


        if (mnproc == 1) then
          write(lp,'(2a)')'RESTART: (mainloop) to file: ',trim(flnmrso)
          write(lp,*)     '     time =                    ',time
          write(lp,*)     '     nstep,iyear,imonth,iday : ',
     &                          nstep,iyear,imonth,iday
          call flush(lp)
        endif

        write(flnm,'(a,i4.4,i2.2,i2.2)')'flxdp.',  iyear,imonth,iday
        !flnmflx = trim(flnmflxd)//trim(flnm)
        !should be done in archive. (flnmflx defined above with flnmarc)!
 
        call restart_out(nstep,dtime, .true.)
 
        call xctmr1(51)

        endif  

      delt1=baclin+baclin

      enddo MAIN   ! end of main time-stepping loop

      call shr_timer_print ( timer_send_to_cpl  )
      call shr_timer_print ( timer_recv_from_cpl)
      call shr_timer_print ( timer_recv_to_send )
      call shr_timer_print ( timer_send_to_recv )


      call ccsm3_exit_HYCOM('hycom normal stop',normal_exit=.true.)

      stop '(hycom normal stop)'
 
!----

      end program hycom
c
c
c> Revision history:
c>
c> May  1997 - removed statement "theta(1)=-thbase" after loop 14
c> June 1997 - added loop 60 to fix bug in vertical summation of -diaflx-
c> Oct. 1999 - option for krt or kpp mixed layer model - convec and diapfl
c>             not called for kpp mixing model
c> Oct. 1999 - dpbl (boundary layer thickness) is output in addition to
c>             dpmixl when the kpp mixing model is selected
c> May  2000 - conversion to SI units
c> Aug. 2000 - added isopycnic (MICOM) vertical coordinate option
c> Oct. 2000 - added option for high frequency atmospheric forcing
c> Nov. 2000 - archive time stamp is either time step or YYYY_DDD_HH
c> Aug. 2002 - added support for multiple tracers
c> Nov. 2002 - more basin-wide surface flux statistics
